package canary

import (
	"time"

	"code.cloudfoundry.org/lager"
)

//go:generate counterfeiter . Chirper
type Chirper interface {
	Chirp(string) (bool, error)
}

//go:generate counterfeiter . Alerter
type Alerter interface {
	NotUnhealthy(time.Time) error
	Unhealthy(time.Time) error
}

//go:generate counterfeiter . StateMachine
type StateMachine interface {
	BecomesUnhealthy(time.Time)
	BecomesNotUnhealthy(time.Time)
}

type CoalMiner struct {
	chirper Chirper
	alerter Alerter
	logger  lager.Logger

	StateMachine StateMachine
}

func NewCoalMiner(chirper Chirper, alerter Alerter, logger lager.Logger) *CoalMiner {
	return &CoalMiner{
		chirper: chirper,
		alerter: alerter,
		logger:  logger,
		StateMachine: &StatefulStateMachine{
			Logger: logger,
			OnBecomesUnhealthy: func(now time.Time) {
				err := alerter.Unhealthy(now)
				if err != nil {
					logger.Error("failure in alerting", err)
				}
			},
			OnBecomesNotUnhealthy: func(now time.Time) {
				err := alerter.NotUnhealthy(now)
				if err != nil {
					logger.Error("failure in alerting", err)
				}
			},
		},
	}
}

func (c *CoalMiner) LetSing(timer <-chan time.Time) {
	for now := range timer {
		ok, err := c.chirper.Chirp(now.String())
		c.logger.Debug("Coalminer transitioning state", lager.Data{"timestamp": now})
		if !ok && err == nil {
			c.logger.Debug(
				"Coalminer received not ok and nil error (replication failure)",
				lager.Data{"timestamp": now},
			)
			c.StateMachine.BecomesUnhealthy(now)
		}
		if ok {
			c.logger.Debug(
				"Coalminer received ok (replication success)",
				lager.Data{"timestamp": now},
			)
			c.StateMachine.BecomesNotUnhealthy(now)
		}
		if !ok && err != nil {
			c.logger.Debug(
				"Coalminer received not ok and non-nil error (cannot determine replication state)",
				lager.Data{"timestamp": now},
			)
			c.StateMachine.BecomesNotUnhealthy(now)
		}

		if err != nil {
			c.logger.Error("failure in chirping", err)
		}
	}
}
