package config

import (
	"flag"

	"github.com/pivotal-cf-experimental/service-config"
	"gopkg.in/validator.v2"

	"code.cloudfoundry.org/cflager"
	"code.cloudfoundry.org/lager"
	"errors"
)

type Config struct {
	PidFile        string `yaml:"PidFile" validate:"nonzero"`
	Logger         lager.Logger
	MySQL          MySQL         `yaml:"MySQL" validate:"nonzero"`
	Canary         Canary        `yaml:"Canary" validate:"nonzero"`
	Notifications  Notifications `yaml:"Notifications" validate:"nonzero"`
	Switchboard    Switchboard   `yaml:"Switchboard" validate:"nonzero"`
	WriteReadDelay int           `yaml:"WriteReadDelay" validate:"nonzero"`
	PollFrequency  int           `yaml:"PollFrequency" validate:"nonzero,min=1"`
	NotifyOnly     bool          `yaml:"NotifyOnly"`
}

type MySQL struct {
	ClusterIPs            []string `yaml:"ClusterIPs" validate:"nonzero"`
	Port                  int      `yaml:"Port" validate:"nonzero"`
	GaleraHealthcheckPort int      `yaml:"GaleraHealthcheckPort" validate:"nonzero"`
}

type Canary struct {
	Database string `yaml:"Database" validate:"nonzero"`
	Username string `yaml:"Username" validate:"nonzero"`
	Password string `yaml:"Password" validate:"nonzero"`
}

type Notifications struct {
	AdminClientUsername string `yaml:"AdminClientUsername" validate:"nonzero"`
	AdminClientSecret   string `yaml:"AdminClientSecret" validate:"nonzero"`
	ClientUsername      string `yaml:"ClientUsername" validate:"nonzero"`
	ClientSecret        string `yaml:"ClientSecret" validate:"nonzero"`
	NotificationsDomain string `yaml:"NotificationsDomain" validate:"nonzero"`
	UAADomain           string `yaml:"UAADomain" validate:"nonzero"`
	ToAddress           string `yaml:"ToAddress" validate:"nonzero"`
}

type Switchboard struct {
	Domains  []string `yaml:"Domains" validate:"nonzero"`
	Username string   `yaml:"Username" validate:"nonzero"`
	Password string   `yaml:"Password" validate:"nonzero"`
}

var InvalidDelay = errors.New("WriteReadDelay must be less than the PollFrequency")

func (c Config) Validate() error {
	if c.PollFrequency <= c.WriteReadDelay {
		return InvalidDelay
	}

	return validator.Validate(c)
}

func NewConfig(osArgs []string) (*Config, error) {
	var rootConfig Config

	binaryName := osArgs[0]
	configurationOptions := osArgs[1:]

	serviceConfig := service_config.New()
	flags := flag.NewFlagSet(binaryName, flag.ExitOnError)

	cflager.AddFlags(flags)

	serviceConfig.AddFlags(flags)
	flags.Parse(configurationOptions)

	rootConfig.Logger, _ = cflager.New(binaryName)

	err := serviceConfig.Read(&rootConfig)

	return &rootConfig, err
}
