package database

import (
	"database/sql"
	"fmt"
	"strings"

	"code.cloudfoundry.org/lager"
)

type Client struct {
	logger lager.Logger
}

func NewClient(logger lager.Logger) *Client {
	return &Client{
		logger: logger,
	}
}

func (c *Client) Setup(db *sql.DB) error {
	_, err := db.Exec("CREATE TABLE IF NOT EXISTS chirps (id INT NOT NULL AUTO_INCREMENT PRIMARY KEY, data VARCHAR(255) NOT NULL) ENGINE=InnoDB")
	if err != nil {
		c.logger.Debug("error creating table", lager.Data{
			"errorMessage": err.Error(),
		})
		return err
	}

	return nil
}

func (c *Client) Write(db *sql.DB, data string) error {
	_, err := db.Exec("INSERT INTO chirps (data) VALUES (?)", data)
	if err != nil {
		c.logger.Debug("error inserting data", lager.Data{
			"errorMessage": err.Error(),
		})
		return err
	}

	return nil
}

// Check returns (true,nil) when the provided data matches the actual data
// Check returns (false,nil) when the provided data does not match the actual data
// Check returns (false,error) if an error occurred while determining whether data matches.
// Check will never return (true,error)
// Check swallows sql.ErrNoRow
func (c *Client) Check(db *sql.DB, data string) (bool, error) {
	var readData string

	err := db.QueryRow("SELECT data FROM chirps WHERE data = ?", data).Scan(&readData)
	if err != nil && err != sql.ErrNoRows {
		c.logger.Debug("error selecting data", lager.Data{
			"errorMessage": err.Error(),
		})
		return false, err
	}

	return (readData == data), nil
}

func (c *Client) Cleanup(db *sql.DB) error {
	// 3 days at once a minute
	numEntriesToKeep := 24 * 60 * 3

	tx, err := db.Begin()
	if err != nil {
		c.logger.Debug("error beginning transaction", lager.Data{
			"errorMessage": err.Error(),
		})
		return err
	}
	defer func() error {
		if err != nil {
			c.logger.Debug("error occurred - rollling back transaction", lager.Data{
				"errorMessage": err.Error(),
			})
			_ = tx.Rollback()
			return err
		}
		return tx.Commit()
	}()

	rows, err := tx.Query("SELECT id FROM chirps ORDER BY id DESC LIMIT ?", numEntriesToKeep)
	if err != nil {
		c.logger.Debug("error selecting rows", lager.Data{
			"errorMessage": err.Error(),
		})

		return err
	}

	// []interface{} because []int does not satisfy []interface{} for use in stmt.Exec(ids...)
	var ids []interface{}

	defer rows.Close()
	for rows.Next() {
		var id int
		err = rows.Scan(&id)
		if err != nil {
			c.logger.Debug("error scanning rows", lager.Data{
				"errorMessage": err.Error(),
			})
			return err
		}

		ids = append(ids, id)
	}

	err = rows.Err()
	if err != nil {
		c.logger.Debug("error on rows", lager.Data{
			"errorMessage": err.Error(),
		})
		return err
	}

	qMarks := strings.Join(strings.Split(strings.Repeat("?", len(ids)), ""), ",")
	query := fmt.Sprintf("DELETE FROM chirps WHERE id NOT IN (%s)", qMarks)

	stmt, err := tx.Prepare(query)
	if err != nil {
		c.logger.Debug("error preparing statement", lager.Data{
			"errorMessage": err.Error(),
		})
		return err
	}

	_, err = stmt.Exec(ids...)
	if err != nil {
		c.logger.Debug("error executing statement", lager.Data{
			"errorMessage": err.Error(),
		})
		return err
	}

	return nil
}
