package database_test

import (
	"errors"

	"code.cloudfoundry.org/lager/lagertest"

	"database/sql"

	"github.com/DATA-DOG/go-sqlmock"
	. "github.com/pivotal-cf-experimental/replication-canary/database"
	"github.com/pivotal-cf-experimental/replication-canary/database/databasefakes"

	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf-experimental/replication-canary/config"
)

var _ = Describe("Client Helper", func() {
	var (
		fakeSwitchboardClient *databasefakes.FakeSwitchboardClient
		testLogger            *lagertest.TestLogger
		clientHelper          *ClientHelper

		dsns []string

		db   *sql.DB
		mock sqlmock.Sqlmock
	)

	BeforeEach(func() {
		dsns = []string{
			"fake_username:fake_password@tcp(fake_ip1:124816)/fake_db",
			"fake_username:fake_password@tcp(fake_ip2:124816)/fake_db",
		}

		fakeSwitchboardClient = &databasefakes.FakeSwitchboardClient{}

		testLogger = lagertest.NewTestLogger("database client helper test")

		config := &config.Config{
			MySQL: config.MySQL{
				ClusterIPs: []string{"fake_ip1", "fake_ip2"},
				Port:       124816,
			},
			Canary: config.Canary{
				Database: "fake_db",
				Username: "fake_username",
				Password: "fake_password",
			},
		}

		clientHelper = NewClientHelperFromConfig(
			config,
			fakeSwitchboardClient,
			testLogger,
		)

		var err error
		db, mock, err = sqlmock.New()
		Expect(err).NotTo(HaveOccurred())
		clientHelper.OpenConn = func(dsn string) (*sql.DB, error) {
			return db, nil
		}
		mock.ExpectExec("^SET SESSION wsrep_sync_wait=1$").WillReturnResult(
			sqlmock.NewResult(1, 1))
		mock.ExpectExec("^SET SESSION wsrep_sync_wait=1$").WillReturnResult(
			sqlmock.NewResult(1, 1))
	})
	AfterEach(func() {
		db.Close()
	})

	Describe("Conns", func() {
		It("sets wsrep_sync_wait on the connections", func() {
			_, err := clientHelper.Conns()
			Expect(err).NotTo(HaveOccurred())
			Expect(mock.ExpectationsWereMet()).NotTo(HaveOccurred())
		})

		Context("when an error occurs setting wsrep_sync_wait", func() {
			It("returns an error", func() {
				db, mock, err := sqlmock.New()
				Expect(err).NotTo(HaveOccurred())
				defer db.Close()
				clientHelper.OpenConn = func(dsn string) (*sql.DB, error) {
					return db, nil
				}
				mock.ExpectExec("^SET SESSION wsrep_sync_wait=1$").WillReturnResult(
					sqlmock.NewResult(1, 1))
				mock.ExpectExec("^SET SESSION wsrep_sync_wait=1$").WillReturnError(
					errors.New("No waiting"))
				_, err = clientHelper.Conns()
				Expect(err).To(HaveOccurred())
				Expect(err).To(MatchError("No waiting"))
				Expect(mock.ExpectationsWereMet()).NotTo(HaveOccurred())
			})
		})
	})

	Describe("Configs", func() {
		It("has the correct DSNs", func() {
			cfgs := clientHelper.MySQLConfigs()
			Expect(cfgs[0].FormatDSN()).To(Equal(dsns[0]))
			Expect(cfgs[1].FormatDSN()).To(Equal(dsns[1]))
		})
	})

	Describe("WriteConn", func() {
		BeforeEach(func() {
			fakeSwitchboardClient.ActiveBackendHostReturns("fake_ip2", nil)
		})

		It("asks the switchboard client for the active backend host", func() {
			_, err := clientHelper.WriteConn()
			Expect(err).NotTo(HaveOccurred())

			Expect(fakeSwitchboardClient.ActiveBackendHostCallCount()).To(Equal(1))
		})

		It("returns the connection that corresponds to the host backend", func() {
			allConns, err := clientHelper.Conns()
			Expect(err).NotTo(HaveOccurred())

			conn, err := clientHelper.WriteConn()
			Expect(err).NotTo(HaveOccurred())

			Expect(conn).To(Equal(allConns[1]))
		})

		Context("when the switchboard client returns an error", func() {
			var (
				expectedErr error
			)

			BeforeEach(func() {
				expectedErr = errors.New("some switchboard error")

				fakeSwitchboardClient.ActiveBackendHostReturns("", expectedErr)
			})

			It("returns the error", func() {
				_, err := clientHelper.WriteConn()
				Expect(err).To(Equal(expectedErr))
			})
		})
	})
})
