package database_test

import (
	dsql "database/sql"
	"errors"

	"code.cloudfoundry.org/lager/lagertest"

	"github.com/DATA-DOG/go-sqlmock"
	. "github.com/onsi/ginkgo"
	. "github.com/onsi/gomega"
	"github.com/pivotal-cf-experimental/replication-canary/database"
)

var _ = Describe("Client", func() {
	var (
		testLogger *lagertest.TestLogger
		client     *database.Client
	)

	BeforeEach(func() {
		testLogger = lagertest.NewTestLogger("database client test")
		client = database.NewClient(testLogger)
	})

	Describe("Setup", func() {
		It("sets up the table if it does not exist", func() {
			db, mock, err := sqlmock.New()
			Expect(err).NotTo(HaveOccurred())

			defer db.Close()

			mock.ExpectExec(`CREATE TABLE IF NOT EXISTS chirps \(id INT NOT NULL AUTO_INCREMENT PRIMARY KEY, data VARCHAR\(255\) NOT NULL\) ENGINE=InnoDB`).WillReturnResult(sqlmock.NewResult(1, 1))

			err = client.Setup(db)
			Expect(err).NotTo(HaveOccurred())

			Expect(mock.ExpectationsWereMet()).NotTo(HaveOccurred())
		})
	})

	Describe("Write", func() {
		It("writes to the connection", func() {
			db, mock, err := sqlmock.New()
			Expect(err).NotTo(HaveOccurred())

			defer db.Close()

			mock.ExpectExec(`INSERT INTO chirps \(data\)`).WithArgs("foobarbaz").WillReturnResult(sqlmock.NewResult(1, 1))

			err = client.Write(db, "foobarbaz")
			Expect(err).NotTo(HaveOccurred())

			Expect(mock.ExpectationsWereMet()).NotTo(HaveOccurred())
		})
	})

	Describe("Check", func() {
		var (
			db   *dsql.DB
			mock sqlmock.Sqlmock
			err  error
		)

		BeforeEach(func() {
			db, mock, err = sqlmock.New()
			Expect(err).NotTo(HaveOccurred())
		})

		AfterEach(func() {
			Expect(mock.ExpectationsWereMet()).NotTo(HaveOccurred())

			db.Close()
		})

		It("verifies the data exists on that connection", func() {
			rows := sqlmock.NewRows([]string{"data"}).AddRow("foobarbaz")
			mock.ExpectQuery(`SELECT data FROM chirps WHERE data =`).WithArgs("foobarbaz").WillReturnRows(rows)

			_, err = client.Check(db, "foobarbaz")
			Expect(err).NotTo(HaveOccurred())
		})

		It("returns error if there's an error in the query", func() {
			dbErr := errors.New("something")

			mock.ExpectQuery(`SELECT data FROM chirps WHERE data =`).WithArgs("foobarbaz").WillReturnError(dbErr)

			_, err = client.Check(db, "foobarbaz")
			Expect(err).To(MatchError(dbErr))
		})

		It("returns false if the row returned does not contain the correct data", func() {
			rows := sqlmock.NewRows([]string{"data"}).AddRow("some-other-data")
			mock.ExpectQuery(`SELECT data FROM chirps WHERE data =`).WithArgs("foobarbaz").WillReturnRows(rows)

			ok, err := client.Check(db, "foobarbaz")
			Expect(err).NotTo(HaveOccurred())
			Expect(ok).To(BeFalse())
		})

		It("swallows sql.ErrNoRow", func() {
			mock.ExpectQuery(`SELECT data FROM chirps WHERE data =`).WithArgs("foobarbaz").WillReturnError(dsql.ErrNoRows)

			ok, err := client.Check(db, "foobarbaz")
			Expect(err).NotTo(HaveOccurred())
			Expect(ok).To(BeFalse())
		})

		It("returns true if the row returned contains the correct data", func() {
			data := "foobarbaz"

			rows := sqlmock.NewRows([]string{"data"}).AddRow(data)
			mock.ExpectQuery(`SELECT data FROM chirps WHERE data =`).WithArgs("foobarbaz").WillReturnRows(rows)

			ok, err := client.Check(db, data)
			Expect(err).NotTo(HaveOccurred())
			Expect(ok).To(BeTrue())
		})
	})

	Describe("Cleanup", func() {
		var (
			db   *dsql.DB
			mock sqlmock.Sqlmock
			err  error
		)

		BeforeEach(func() {
			db, mock, err = sqlmock.New()
			Expect(err).NotTo(HaveOccurred())
		})

		AfterEach(func() {
			Expect(mock.ExpectationsWereMet()).NotTo(HaveOccurred())

			db.Close()
		})

		It("deletes any entries not in the most recent 4,320", func() {
			rows := sqlmock.NewRows([]string{"id"}).AddRow(1).AddRow(2).AddRow(3)

			mock.ExpectBegin()
			mock.ExpectQuery("SELECT id FROM chirps ORDER BY id DESC LIMIT").WithArgs(4320).WillReturnRows(rows)

			// Three question marks for three returned rows
			prepare := mock.ExpectPrepare(`DELETE FROM chirps WHERE id NOT IN \(\?,\?,\?\)`)
			prepare.ExpectExec().WithArgs(1, 2, 3).WillReturnResult(sqlmock.NewResult(0, 0))

			mock.ExpectCommit()

			err := client.Cleanup(db)

			Expect(err).NotTo(HaveOccurred())
		})

		It("performs a transaction rollback if there is an error", func() {
			mock.ExpectBegin()
			mock.ExpectQuery("SELECT id FROM chirps ORDER BY id DESC LIMIT").WithArgs(4320).WillReturnError(errors.New("some exec error"))
			mock.ExpectRollback()

			err := client.Cleanup(db)

			Expect(err).To(HaveOccurred())
		})

		It("returns error when creating transaction fails", func() {
			expectedErr := errors.New("some error")

			mock.ExpectBegin().WillReturnError(expectedErr)

			err := client.Cleanup(db)
			Expect(err).To(Equal(expectedErr))
		})

		It("returns error when select query fails", func() {
			expectedErr := errors.New("some error")

			mock.ExpectBegin()
			mock.ExpectQuery("SELECT id FROM chirps ORDER BY id DESC LIMIT").WithArgs(4320).WillReturnError(expectedErr)
			mock.ExpectRollback()

			err := client.Cleanup(db)
			Expect(err).To(Equal(expectedErr))
		})

		It("returns error when scanning result of select fails", func() {
			rows := sqlmock.NewRows([]string{"id"}).AddRow("string")

			mock.ExpectBegin()
			mock.ExpectQuery("SELECT id FROM chirps ORDER BY id DESC LIMIT").WithArgs(4320).WillReturnRows(rows)
			mock.ExpectRollback()

			err := client.Cleanup(db)
			Expect(err).To(HaveOccurred())
		})

		It("returns error when rows.Err fails", func() {
			expectedErr := errors.New("some error")
			rows := sqlmock.NewRows([]string{"id"}).AddRow(1)

			mock.ExpectBegin()
			mock.ExpectQuery("SELECT id FROM chirps ORDER BY id DESC LIMIT").WithArgs(4320).WillReturnRows(rows)
			rows.RowError(0, expectedErr)
			mock.ExpectRollback()

			err := client.Cleanup(db)
			Expect(err).To(Equal(expectedErr))
		})

		It("returns error when preparing delete query fails", func() {
			expectedErr := errors.New("some error")
			rows := sqlmock.NewRows([]string{"id"}).AddRow(1)

			mock.ExpectBegin()
			mock.ExpectQuery("SELECT id FROM chirps ORDER BY id DESC LIMIT").WithArgs(4320).WillReturnRows(rows)
			mock.ExpectPrepare("DELETE FROM chirps WHERE id NOT IN").WillReturnError(expectedErr)
			mock.ExpectRollback()

			err := client.Cleanup(db)
			Expect(err).To(Equal(expectedErr))
		})

		It("returns error when executing delete query fails", func() {
			expectedErr := errors.New("some error")
			rows := sqlmock.NewRows([]string{"id"}).AddRow(1)

			mock.ExpectBegin()
			mock.ExpectQuery("SELECT id FROM chirps ORDER BY id DESC LIMIT").WithArgs(4320).WillReturnRows(rows)
			prepare := mock.ExpectPrepare("DELETE FROM chirps WHERE id NOT IN")
			prepare.ExpectExec().WithArgs(1).WillReturnError(expectedErr)
			mock.ExpectRollback()

			err := client.Cleanup(db)
			Expect(err).To(Equal(expectedErr))
		})
	})
})
