package main

import (
	"os"

	"io/ioutil"
	"strconv"
	"time"

	"code.cloudfoundry.org/clock"
	"code.cloudfoundry.org/lager"
	"github.com/cloudfoundry-incubator/uaa-go-client"
	uaa_config "github.com/cloudfoundry-incubator/uaa-go-client/config"
	"github.com/pivotal-cf-experimental/notifications-client/notificationemailer"
	"github.com/pivotal-cf-experimental/replication-canary/alert"
	"github.com/pivotal-cf-experimental/replication-canary/canary"
	"github.com/pivotal-cf-experimental/replication-canary/clientcreator"
	"github.com/pivotal-cf-experimental/replication-canary/config"
	"github.com/pivotal-cf-experimental/replication-canary/database"
	"github.com/pivotal-cf-experimental/replication-canary/galera"
	"github.com/pivotal-cf-experimental/replication-canary/switchboard"
)

func main() {
	appConfig, err := config.NewConfig(os.Args)

	logger := appConfig.Logger

	if err != nil {
		logger.Fatal("Failed to read config", err, lager.Data{
			"config": sanitizeConfig(appConfig),
		})
	}

	err = appConfig.Validate()
	if err != nil {
		logger.Fatal("Failed to validate config", err, lager.Data{
			"config": sanitizeConfig(appConfig),
		})
	}

	pidfile, err := os.Create(appConfig.PidFile)
	if err != nil {
		logger.Fatal("Failed to create a file", err)
	}

	ioutil.WriteFile(pidfile.Name(), []byte(strconv.Itoa(os.Getpid())), 0644)

	logger.Info("Starting replication canary with configuration", lager.Data{
		"mysql": appConfig.MySQL,
	})

	notificationsClient := &notificationemailer.Client{
		NotificationsDomain: "https://" + appConfig.Notifications.NotificationsDomain,
		SkipSSLCertVerify:   true,
	}

	adminUaaConfig := &uaa_config.Config{
		ClientName:       appConfig.Notifications.AdminClientUsername,
		ClientSecret:     appConfig.Notifications.AdminClientSecret,
		UaaEndpoint:      "https://" + appConfig.Notifications.UAADomain,
		SkipVerification: true,
	}

	adminUaaClient, err := uaa_go_client.NewClient(
		logger,
		adminUaaConfig,
		clock.NewClock(),
	)
	if err != nil {
		logger.Fatal("Failed to generate a UAA client", err)
	}

	err = clientcreator.CreateClient(adminUaaClient, appConfig)
	if err != nil {
		logger.Fatal("Failed to register UAA client", err)
	}

	uaaConfig := &uaa_config.Config{
		ClientName:       appConfig.Notifications.ClientUsername,
		ClientSecret:     appConfig.Notifications.ClientSecret,
		UaaEndpoint:      "https://" + appConfig.Notifications.UAADomain,
		SkipVerification: true,
	}

	uaaClient, err := uaa_go_client.NewClient(
		logger,
		uaaConfig,
		clock.NewClock(),
	)
	if err != nil {
		logger.Fatal("Failed to generate a UAA client", err)
	}

	loggingAlerter := &alert.LoggingAlerter{
		Logger: logger,
	}

	emailingAlerter := &alert.EmailingAlerter{
		Logger:              logger,
		UAAClient:           uaaClient,
		NotificationsClient: notificationsClient,
		ToAddress:           appConfig.Notifications.ToAddress,
	}

	aggregateAlerter := alert.AggregateAlerter{
		loggingAlerter,
		emailingAlerter,
	}

	var sc *switchboard.Client

	skipSSLCertVerify := true
	for i, d := range appConfig.Switchboard.Domains {
		switchboardClient := switchboard.NewClient(
			d,
			appConfig.Switchboard.Username,
			appConfig.Switchboard.Password,
			skipSSLCertVerify,
			logger,
		)

		switchboardAlerter := &alert.SwitchboardAlerter{
			Logger:            logger,
			SwitchboardClient: switchboardClient,
			NoOp:              appConfig.NotifyOnly,
		}

		aggregateAlerter = append(aggregateAlerter, switchboardAlerter)

		if i == 0 {
			sc = switchboardClient
		}
	}

	helper := database.NewClientHelperFromConfig(appConfig, sc, logger)
	client := database.NewClient(logger)
	healthchecker := &galera.Client{
		Logger: logger,
	}
	bird := canary.NewCanary(client, helper, healthchecker, time.Duration(appConfig.WriteReadDelay)*time.Second, logger)

	err = bird.Setup()
	if err != nil {
		logger.Fatal("Canary setup failed", err)
	}

	coalMiner := canary.NewCoalMiner(
		bird,
		aggregateAlerter,
		logger,
	)

	ticker := time.NewTicker(time.Duration(appConfig.PollFrequency) * time.Second)
	logger.Info("ready to sing")
	coalMiner.LetSing(ticker.C)
}

func sanitizeConfig(cfg *config.Config) *config.Config {
	cfg.Canary.Password = "REDACTED"

	return cfg
}
