package notificationemailer

import (
	"bytes"
	"crypto/tls"
	"encoding/json"
	"fmt"
	"io/ioutil"
	"net/http"
	"time"
)

type Client struct {
	NotificationsDomain string
	SkipSSLCertVerify   bool
}

type notificationsEmailRequestBody struct {
	To      string `json:"to"`
	Subject string `json:"subject"`
	HTML    string `json:"html"`
	KindID  string `json:"kind_id"`
}

func (c *Client) Email(clientToken, to, subject, html, kindID string) error {
	body, err := json.Marshal(notificationsEmailRequestBody{
		To:      to,
		Subject: subject,
		HTML:    html,
		KindID:  kindID,
	})
	if err != nil {
		// Untested because we cannot get Marshal to fail - none of the fields
		// are un-marshallable
		return err
	}

	url := c.NotificationsDomain + "/emails"
	req, err := http.NewRequest("POST", url, bytes.NewReader(body))
	if err != nil {
		return err
	}

	req.Header.Add("Content-Type", "application/x-www-form-urlencoded")
	req.Header.Add("X-NOTIFICATIONS-VERSION", "1")
	req.Header.Add("Authorization", "Bearer "+clientToken)

	client := http.Client{
		Timeout: time.Second * 5,
		Transport: &http.Transport{
			TLSClientConfig: &tls.Config{
				InsecureSkipVerify: c.SkipSSLCertVerify,
			},
		},
	}
	res, err := client.Do(req)
	if err != nil {
		return err
	}
	defer res.Body.Close()

	if res.StatusCode >= http.StatusBadRequest {
		b, _ := ioutil.ReadAll(res.Body)

		return fmt.Errorf("bad response %s (%d) - %s", "sending email", res.StatusCode, string(b))
	}

	return nil
}
